@@ -1,7 +1,7 @@
 """
 
-m2wsgi:  a mongrel2 => wsgi gateway
-===================================
+m2wsgi:  a mongrel2 => wsgi gateway and helper tools
+====================================================
 
 
 This module provides a WSGI gateway handler for the Mongrel2 webserver,
@@ -13,7 +13,6 @@ You might also find its supporting classes useful for developing non-WSGI
 handlers in python.
 
 
-
 Command-line usage
 ------------------
 
@@ -45,7 +44,7 @@ If you have more complicated needs, you can use m2wsgi from within your
 application.  The main class is 'WSGIHandler' which provides a simple
 server interface.  The equivalent of the above command-line usage is::
 
-    from m2wsgi.base import WSGIHandler
+    from m2wsgi.io.standard import WSGIHandler
     handler = WSGIHandler(my_wsgi_app,"tcp://127.0.0.1:9999")
     handler.serve()
 
@@ -54,12 +53,12 @@ that the Mongrel2 recv socket is on the next port down from the send socket,
 and that it's OK to connect the socket without a persistent identity.
 
 For finer control over the connection between your handler and Mongrel2,
-create your own Connection object::
+create your own Connection object.  Here we use 127.0.0.1:9999 as the send
+socket with identity AA-BB-CC, and use 127.0.0.2:9992 as the recv socket::
 
-    from m2wsgi.base import WSGIHandler, Connection
-    conn = Connection(send_spec="tcp://127.0.0.1:9999",
-                      recv_spec="tcp://127.0.0.1:9992",
-                      send_ident="9a5eee79-dbd5-4f33-8fd0-69b304c6035a")
+    from m2wsgi.io.standard import WSGIHandler, Connection
+    conn = Connection(send_sock="tcp://AA-BB-CC@127.0.0.1:9999",
+                      recv_sock="tcp://127.0.0.1:9992")
     handler = WSGIHandler(my_wsgi_app,conn)
     handler.serve()
 
@@ -67,11 +66,56 @@ If you're creating non-WSGI handlers for Mongrel2 you might find the following
 classes useful:
 
     :Connection:  represents the connection from your handler to Mongrel2,
-                  through which you can read requests and rend responses.
+                  through which you can read requests and send responses.
+
+    :Client:      represents a client connected to the server, to whom you
+                  can send data at any time.
 
     :Request:     represents a client request to which you can asynchronously
                   send response data at any time.
 
+    :Handler:     a base class for implementing handlers, with nothing
+                  WSGI-specific in it.
+
+
+Middleware
+----------
+
+If you need to add fancy features to the server, you can specify additional
+WSGI middleware that should be applied around the application.  For example,
+m2wsgi provides a gzip-encoding middleware that can be used to compress
+response data::
+
+    m2wsgi --middleware=GZipMiddleware
+           dotted.app.name tcp://127.0.0.1:9999
+
+If you want additional compression at the expense of WSGI compliance, you
+can also do some in-server buffering before the gzipping is applied:
+
+    m2wsgi --middleware=GZipMiddleware
+           --middleware=BufferMiddleware
+           dotted.app.name tcp://127.0.0.1:9999
+
+The default module for loading middleware is m2wsgi.middleware; specify a
+full dotted name to load a middleware class from another module.
+
+
+Devices
+-------
+
+This module also provides a number of pre-built "devices" - stand-alone
+executables designed to perform a specific common task.  Currently availble
+devices are:
+
+    :dispatcher:  implements a more flexible request-routing scheme than
+                  the standard mongrel2 PUSH socket.
+
+    :response:    implements a simple canned response, with ability to
+                  interpolate variables from the request.
+
+    :reaper:      a device for timing out hung or inactive requests.
+
+
 
 Don't we already have one of these?
 -----------------------------------
@@ -97,9 +141,10 @@ as you might find in e.g. the CherryPy server.  Instead, you just start up
 as many threads as you need, have them all connect to the same handler socket,
 and mongrel2 (via zmq) will automatically load-balance the requests to them.
 
-Similarly, there's no explicit support for reloading when the code changes.
-Just kill the old handler and start up a new one.  If you're using fixed
-handler UUIDs then zmq will ensure that the handover happens gracefully.
+Similarly, there's no fancy arrangement of master/worker processes to support
+clean reloading of the handler; you just kill the old handler process and start
+up a new one.  Send m2wsgi a SIGHUP and it will automatically shutdown and
+reincarnate itself for a clean restart.
 
 
 Current bugs, limitations and things to do
@@ -110,32 +155,27 @@ It's not all perfect just yet, although it does seem to mostly work:
     * Needs tests something fierce!  I just have to find the patience to
       write the necessary setup and teardown cruft.
 
-    * When running multiple threads, ctrl-C doesn't cleanly exit the process.
-      Seems like the background threads get stuck in a blocking recv().
-      I *really* don't want to emulate interrupts using zmq_poll...
-
-    * The zmq load-balancing algorithm is greedy round-robin, which isn't
-      ideal.  For example, it can schedule several fast requests to the same
-      thread as a slow one, making them wait even if other threads become
-      available.  I'm working on a zmq adapter that can do something better
-      (see the pull2xreq device in this distribution).
-
     * It would be great to grab connection details straight from the
       mongrel2 config database.  Perhaps a Connection.from_config method
       with keywords to select the connection by handler id, host, route etc.
 
+    * support for except-100-continue; this may have to live in mongrel2
 
 """
+#  Copyright (c) 2011, Ryan Kelly.
+#  All rights reserved; available under the terms of the MIT License.
 
 __ver_major__ = 0
-__ver_minor__ = 3
+__ver_minor__ = 4
 __ver_patch__ = 0
 __ver_sub__ = ""
 __version__ = "%d.%d.%d%s" % (__ver_major__,__ver_minor__,__ver_patch__,__ver_sub__)
 
 
 import sys
+import os
 import optparse
+from subprocess import MAXFD
 from textwrap import dedent
 try:
     import signal
@@ -151,70 +191,87 @@ def main(argv=None):
     op = optparse.OptionParser(usage=dedent("""
     usage:  m2wsgi [options] dotted.app.name spend_spec [recv_spec]
     """))
-    op.add_option("","--io",default="base",
+    op.add_option("","--io",default="standard",
                   help="the I/O module to use")
     op.add_option("","--num-threads",type="int",default=1,
                   help="the number of threads to use")
-    op.add_option("","--send-ident",type="str",default=None,
-                  help="the send socket identity to use")
-    op.add_option("","--recv-ident",type="str",default=None,
-                  help="the recv socket identity to use")
-    op.add_option("","--send-type",type="str",default=None,
-                  help="the send socket type to use")
-    op.add_option("","--recv-type",type="str",default=None,
-                  help="the recv socket type to use")
+    op.add_option("","--conn-type",default="",
+                  help="the type of connection to use")
+    op.add_option("","--middleware",action="append",
+                  help="any middleware to apply to the wsgi app")
     (opts,args) = op.parse_args(argv)
     #  Sanity-check the arguments.
     if len(args) < 1:
         raise ValueError("you must specify the WSGI application")
-    if len(args) < 2:
-        raise ValueError("you must specify the mongrel2 request socket")
-    if len(args) > 3:
-        raise ValueError("too many arguments")
     if opts.num_threads <= 0:
         raise ValueError("--num-threads must be positive")
-    #  Grab the application, connection and handler class
+    #  Grab the connection and handler class from the IO module
+    conn_args = args[1:]
+    iomod = "%s.io.%s" % (__name__,opts.io,)
+    iomod = __import__(iomod,fromlist=["WSGIHandler"])
+    iomod.monkey_patch()
+    WSGIHandler = iomod.WSGIHandler
+    if not opts.conn_type:
+        Connection = iomod.Connection
+    else:
+        for nm in dir(iomod):
+            if nm.lower() == opts.conn_type.lower() + "connection":
+                Connection = getattr(iomod,nm)
+                break
+        else:
+            raise ValueError("unknown connection type: %s" % (opts.conn_type,))
+    #  Now that things are monkey-patched, we can load the app
+    #  and the threading module.
+    import threading
     app = load_dotted_name(args[0])
     assert callable(app), "the specified app is not callable"
-    conn_args = args[1:]
-    conn_kwds = dict(send_ident=opts.send_ident,
-                     recv_ident=opts.recv_ident,
-                     send_type=opts.send_type,
-                     recv_type=opts.recv_type)
-    try:
-        iomod = "%s.%s" % (__name__,opts.io,)
-        iomod = __import__(iomod,fromlist=["WSGIHandler"])
-    except (ImportError,AttributeError,):
-        raise
-        raise ValueError("not a m2wsgi IO module: %r" % (opts.io,))
-    if hasattr(iomod,"monkey_patch"):
-        iomod.monkey_patch()
-    #  Try hard to clean up properly when killed
+    #  Apply any middleware that was specified in the args
+    if opts.middleware:
+       for mname in reversed(opts.middleware):
+           if "." not in mname:
+               mname = "m2wsgi.middleware." + mname
+           mcls = load_dotted_name(mname)
+           app = mcls(app)
+    #  Try to clean up properly when killed.
+    #  We turn SIGTERM into a KeyboardInterrupt exception.
+    #  We catch SIGHUP and re-execute ourself as a simple reload mechanism.
+    reload_the_process = []
     if signal is not None:
-        def interrupt():
+        def on_sigterm(*args):
             raise KeyboardInterrupt
-        signal.signal(signal.SIGTERM,lambda *a: interrupt)
+        signal.signal(signal.SIGTERM,on_sigterm)
+        def on_sighup(*args):
+            reload_the_process.append(True)
+            raise KeyboardInterrupt
+        signal.signal(signal.SIGHUP,on_sighup)
     #  Start the requested N handler threads.
     #  N-1 are started in background threads, then one on this thread.
     handlers = []
     threads = []
     def run_handler():
-        conn = iomod.WSGIHandler.ConnectionClass(*conn_args,**conn_kwds)
-        handler = iomod.WSGIHandler(app,conn)
+        conn = Connection(*conn_args)
+        handler = WSGIHandler(app,conn)
         handlers.append(handler)
         handler.serve()
-    import threading
     for i in xrange(opts.num_threads - 1):
         t = threading.Thread(target=run_handler)
         t.start()
         threads.append(t)
     try:
         run_handler()
+    except KeyboardInterrupt:
+        if not reload_the_process:
+            raise
     finally:
         #  When the handler exits, shut down any background threads.
         for h in handlers:
             h.stop()
         for t in threads:
             t.join()
+    #  If we're doing a clean restart, close all fds and exec ourself.
+    if reload_the_process:
+        Connection.ZMQ_CTX.term()
+        os.closerange(3,MAXFD)
+        os.execv(sys.argv[0],sys.argv)
 
 
